/* This module allows proxy authentication via the TIS authserver.
   The authserver can be configured to use passwords, but in most
   cases it uses one of several challenge/response schemes. HTTP Basic
   Authentication does not support a challenge which is displayed to
   the user, so we use a trick here: put the challenge into the realm
   parameter (which unfortunately requires some changes in the core).

   The user then goes on like this: when first seeing the dialog
   "Proxy authentication required for Squid proxy-caching
   webserver...", enter your user name and an empty password.
   Then you will get "Proxy authentication failed". Push "Retry".
   Now the challenge is in the dialog, like "Proxy authentication
   required for S/Key challenge..." Enter the correct response as
   password.

   This was verified to work with Netscape 4.5 and Lynx 2.8, but no
   browser which supports authentication at all should be stupid
   enough to not display the actual realm parameter or not give the
   user a chance to retry.

   Olaf Titz, 1999-09-06
*/

#include "squid.h"
#include "module.h"

#define MODNAME "mod_auth_authsrv"

classdef(private authsrvObject, moduleObject, ({
    char *host;
    int port;
}))

#define AUTH_AUTHSRV_C
#include <classes.dh>

Ddef(authsrvObject)
{
    xfree(leakFree(this->host));
}

typedef enum {
    AS_CONNECT, AS_CHALL, AS_RES
} asState;

#define TIMEOUT 15      /* authsrv connection timeout */
#define SMBUF   256     /* buffer size */

typedef struct {
    asState state;
    int fd;
    char *result;
    moduleAuthCBData *data;
    authsrvObject *owner;
    char buf[SMBUF];
    int pos;
} asCBData;

static CNCB auth_authsrvCCB;
static PF auth_authsrvRCB;
static CWCB auth_authsrvWCB;
static PF auth_authsrvClose;
static PF auth_authsrvTimeout;

static void cbdataLfree(void *p, int n)
{
    cbdataXfree(leakFree(p), n);
}

/* "Filter" function.
   Conect to the authsrv; at the end will call arg->authResult()
   (via the close handler).
*/
static void *auth_authsrvFilter(authsrvObject *this, const void *arg)
{
    moduleAuthCBData *m = (moduleAuthCBData*)arg;
    int fd;
    struct in_addr me;
    asCBData *a;

    me.s_addr = INADDR_ANY;
    fd = comm_open(SOCK_STREAM, 0, me, 0, COMM_NONBLOCKING, "authsrv");
    if (fd == COMM_ERROR) {
        debug(95, 1) (MODNAME ": comm_open failed\n");
        m->authResult(m, RREJECT);
        return NULL;
    }
    a = leakAdd(xmalloc(sizeof(asCBData)));
    cbdataAdd(a, cbdataLfree, 0);
    cbdataLock(a);
    cbdataLock(m);
    a->state = AS_CONNECT;
    a->fd = fd;
    a->result = RREJECT;
    a->data = m;
    a->owner = REF(this);
    a->pos = 0;
    comm_add_close_handler(fd, auth_authsrvClose, a);
    commSetTimeout(fd, TIMEOUT, auth_authsrvTimeout, a);
    commConnectStart(fd, this->host, this->port, auth_authsrvCCB, a);
    return NULL;
}

/* Connection complete handler. */
static void auth_authsrvCCB(int fd, int status, void *data)
{
    if (status != COMM_OK) {
        debug(95, 1) (MODNAME ": connect: %d\n", status);
        comm_close(fd);
        return;
    }
    commSetTimeout(fd, TIMEOUT, auth_authsrvTimeout, data);
    commSetSelect(fd, COMM_SELECT_READ, auth_authsrvRCB, data, 0);
}

/* Read handler.
   Here we do the real protocol work.
*/
static void auth_authsrvRCB(int fd, void *data)
{
    asCBData *a = data;
    int len = -1, i;
    MemBuf mb;

    Counter.syscalls.sock.reads++;
    len = read(fd, a->buf + a->pos, SMBUF - a->pos);
    fd_bytes(fd, len, FD_READ);
    if (len <= 0) {
        debug(95, 3) (MODNAME ": read: %d\n", len ? errno : len);
	comm_close(fd);
	return;
    }
    /* Read a whole line? If not, continue. */
    for (i = a->pos; i < a->pos + len; ++i)
        if (a->buf[i] == '\n') {
            a->buf[i] = '\0';
            break;
        }
    if (i >= a->pos + len) {
        a->pos += len;
        if (a->pos >= SMBUF) {
            /* authserv fed us garbage */
            debug(95, 1) (MODNAME ": buffer overrun\n");
            comm_close(fd);
            return;
        }
        commSetSelect(fd, COMM_SELECT_READ, auth_authsrvRCB, data, 0);
        return;
    }
    a->pos = 0;
    debug(95, 8) (MODNAME ": %d<%s>\n", a->state, a->buf);

    switch (a->state) {

    case AS_CONNECT:
        if (strncmp(a->buf, "Authsrv ready", 13)) {
            debug(95, 1) (MODNAME ": authsrv(ready): %s\n", a->buf);
            comm_close(fd);
            return;
        }
        memBufDefInit(&mb);
        memBufPrintf(&mb, "authorize %s squid\n",
                     a->data->check->auth_user->user);
        a->state = AS_CHALL;
        comm_write_mbuf(fd, mb, auth_authsrvWCB, a);
        return;

    case AS_CHALL:
        if (!strncmp(a->buf, "challenge ", 10)) {
            a->result = xstrdup(a->buf+10);
            goto resp;
        }
        if (strncmp(a->buf, "password", 8)) {
            debug(95, 1) (MODNAME ": authserver(challenge): %s\n", a->buf);
            comm_close(fd);
            return;
        }
    resp:
        memBufDefInit(&mb);
        memBufPrintf(&mb, "response '%s'\n",
                     a->data->check->auth_user->passwd);
        a->state = AS_RES;
        comm_write_mbuf(fd, mb, auth_authsrvWCB, a);
        return;

    case AS_RES:
        if (strncmp(a->buf, "ok", 2))
            debug(95, 3) (MODNAME ": authserver(password): %s\n", a->buf);
        else
            a->result = NULL; /* Pass! */
        comm_close(fd);
    }
}

/* Write handler.
   Just keep the ball rolling.
*/
static void auth_authsrvWCB(int fd, char *buf, size_t size,
                            int err, void *data)
{
    if (err) {
        debug(95, 1) (MODNAME ": write: %d\n", err);
        comm_close(fd);
        return;
    }
    commSetTimeout(fd, TIMEOUT, auth_authsrvTimeout, data);
    commSetSelect(fd, COMM_SELECT_READ, auth_authsrvRCB, data, 0);
}

/* Connection timeout handler.
*/
static void auth_authsrvTimeout(int fd, void *data)
{
    debug(95, 1) (MODNAME ": timeout\n");
    comm_close(fd);
}

/* Connection close handler.
   This calls the result callback.
*/
static void auth_authsrvClose(int fd, void *data)
{
    asCBData *a = data;
    moduleAuthCBData *m = a->data;
    comm_close(a->fd);
    if (cbdataValid(m))
        m->authResult(m, a->result);
    cbdataUnlock(m);
    UNREF(a->owner);
    cbdataUnlock(a);
    cbdataFree(a);
}

void moduleInit(const wordlist *args)
{
    authsrvObject *m;
    int p;

    if (!args || !args->next) {
        debug(95, 1) (MODNAME ": needs two arguments\n");
        return;
    }
    if (!(p = atoi(args->next->key))) {
        debug(95, 1) (MODNAME ": argument syntax error\n");
        return;
    }
    m = new(authsrvObject);
    m->chain.typ = FIL_AUTH;
    m->info.version = MODULE_API_VERSION;
    m->description = "TIS FWTK authsrv authenticator";
    m->filter = auth_authsrvFilter;
    m->host = leakAdd(xstrdup(args->key));
    m->port = p;

    moduleRegister((moduleObject *)m);
}
