/* Password-file based authentication module.
   This module takes as argument the name of a password file, default
   is "/etc/passwd". The file is read in completely whenever it changes,
   lookup is done in memory.

   We don't use getpwnam() because that could block (e.g. when using
   NIS), and thus would block the whole Squid process. If the passwd
   file is to be taken from NIS, "ypcat" a local copy.

   Note: empty passwords in the file mean "match none" instead of
   "match all" here.

   Olaf Titz, 1999-09-07
*/

#include "squid.h"
#include "module.h"
#ifdef HAVE_CRYPT_H
#include "crypt.h"
#endif

#define MODNAME "mod_auth_passwd"

#undef DEBUG_HTABLE /* for debugging the hash tables */

typedef struct _passwdList {    /* Structure for storing one passwd line */
    struct _passwdList *next;           /* chain */
    char *user;                         /* user name */
    char *passwd;                       /* crypt()ed password */
} passwdList;

classdef(private passwdObject, moduleObject, ({
    char *fileName;                     /* Password file name */
    time_t mtime;                       /* Password file mtime */
    float sizefactor;                   /* Relative hash table size */
    int buckets;                        /* Actual hash table size */
    passwdList **pwd;                   /* The hash table */
}))

#define AUTH_PASSWD_C
#include <classes.dh>

static void passwdUnload(passwdObject *d);

Ddef(passwdObject)
{
    passwdUnload(this);
    xfree(leakFree(this->fileName));
}

#ifdef DEBUG_HTABLE
/* Dump the hash table */
static void passwd_dump_table(passwdObject *d)
{
    int i;
    passwdList *w;
    char buf[128];
    MemBuf mb;
    memBufDefInit(&mb);
    for (i=0; i<d->buckets; ++i)
        if (d->pwd[i]) {
            snprintf(buf, sizeof(buf), "%d", i);
            memBufAppend(&mb, buf, strlen(buf));
            for (w=d->pwd[i]; w; w=w->next) {
                snprintf(buf, sizeof(buf), "\t%s\t%s\n", w->user, w->passwd);
                memBufAppend(&mb, buf, strlen(buf));
            }
        }
    memBufAppend(&mb, "\0", 1);
    debug(95, 9) ("auth_passwd.c: %s: %d\n%s", d->fileName, d->buckets,
                  mb.buf);
    memBufClean(&mb);
}
#endif

/* Free the in-memory copy of password file */
static void passwdUnload(passwdObject *d)
{
    passwdList *w, *w0;
    int i;
    for (i=0; i<d->buckets; ++i) {
        for (w = d->pwd[i]; w; w = w0) {
            w0 = w->next;
            xfree(leakFree(w->user));
            xfree(leakFree(w->passwd));
            xfree(leakFree(w));
        }
    }
    if (d->buckets)
        xfree(leakFree(d->pwd));
    d->buckets = 0;
}

/* String hash function. */
static unsigned int simpleHash(const unsigned char *p, int n)
{
    register unsigned int r = 0;
    register unsigned char c;
    while ((c = *p++))
        r = ((r<<19) | (r>>13)) ^ c;
    return r % n;
}

/* Euclid's. */
static unsigned int gcd(unsigned int a, unsigned int b)
{
    while (1) {
        if (!a)
            return b;
        if (!b)
            return a;
        if (a > b)
            a %= b;
        else
            b %= a;
    }
}

/* Reload the password file, if necessary */
static void passwdCheckReload(passwdObject *d)
{
    struct stat st;
    FILE *f;
    char buf[256];
    passwdList *w0 = NULL, *w1 = NULL, *w;
    char *p1, *p2;
    int ln = 0, n, z0 = 0, z1 = 0;

    if (stat(d->fileName, &st)<0) {
        debug(95, 1) ("%s: stat: %s\n", d->fileName, xstrerror());
        return;
    }
    if (d->mtime >= st.st_mtime)
        return;
    d->mtime = st.st_mtime;

    f = fopen(d->fileName, "r");
    if (!f) {
        debug(95, 1) ("%s: open: %s\n", d->fileName, xstrerror());
        return;
    }
    debug(95, 4) (MODNAME ": reloading %s\n", d->fileName);
    passwdUnload(d);

    /* Read all lines of password file into a single chain */
    while (fgets(buf, sizeof(buf), f)) {
        for (p1=buf; *p1 && *p1!=':'; ++p1);
        if (*p1!=':')
            continue;
        *p1++='\0';
        if (*p1=='!' || *p1=='*' || *p1==':' || *p1=='\n')
            continue;
        for (p2=p1; *p2 && *p2!=':' && *p2!='\n'; ++p2);
        *p2='\0';
        w = leakAdd(xmalloc(sizeof(passwdList)));
        w->user = leakAdd(xstrdup(buf));
        w->passwd = leakAdd(xstrdup(p1));
        if (w1)
            w1->next = w;
        else
            w0 = w;
        w1 = w;
        ++ln;
    }
    if (!ln)
        return;
    w1->next = NULL;
    fclose(f);
    debug(95, 4) (MODNAME ": loaded %d users\n", ln);

    /* Allocate hash table. Size is odd, >=5 and relatively prime to ln. */
    n = ((int)(d->sizefactor * ln) + 4) | 1;
    while (gcd(ln, n) > 1)
        n += 2;
    d->buckets = n;
    d->pwd = leakAdd(xcalloc(d->buckets, sizeof(passwdList *)));

    /* Distribute the passwd entries into the hash table */
    for (w = w0; w; w = w1) {
        z1+=2;
        n = simpleHash(w->user, d->buckets);
        if (d->pwd[n]) {
            ++z1;
            for (w1 = d->pwd[n]; w1->next; w1 = w1->next, ++z1);
            w1->next = w;
        } else {
            d->pwd[n] = w;
            ++z0;
        }
        w1 = w->next;
        w->next = NULL;
    }
    /* The numbers presented here are: allocated buckets, used buckets,
       average string comparisons needed to find an entry. The latter is at
       least 2.0 because the initial hash calculation is just as expensive
       as one comparison.
    */
    debug(95, 4) (MODNAME ": hashtable usage %d %d %.3f\n", d->buckets,
                  z0, (float)z1/(float)ln);
#ifdef DEBUG_HTABLE
    if (debugLevels[95]>=9)
        passwd_dump_table(d);
#endif
}


/* moduleObject->filter for FIL_AUTH:
   arg is moduleAuthCBData struct which contains an aclCheck_t which in
   turn contains an acl_proxy_auth_user with user and passwd.
   Calls or enqueues call to arg->authResult(arg, result) with result as:
      NULL              if auth succeeded
      RREJECT           if auth failed
      alloced string    for presenting a challenge
   Return value is ignored.
*/
static void *auth_passwdFilter(passwdObject *this, const void *arg)
{
    moduleAuthCBData *m = (moduleAuthCBData*)arg;
    const acl_proxy_auth_user *a = m->check->auth_user;

    passwdCheckReload(this);
    if (this->buckets) {
        passwdList *w;
        for (w = this->pwd[simpleHash(a->user, this->buckets)]; w; w = w->next)
            if (!strcmp(w->user, a->user)) {
                if (!strcmp(w->passwd, (char*)crypt(a->passwd, w->passwd))) {
                    debug(95, 3) (MODNAME ": user `%s' pass\n", a->user);
                    m->authResult(m, NULL);
                    return NULL;
                } else {
                    debug(95, 3) (MODNAME ": user `%s' password mismatch\n",
                                  a->user);
                }
            }
    }
    debug(95, 3) (MODNAME ": user `%s' not found\n", a->user);
    m->authResult(m, RREJECT);
    return NULL;
}

void moduleInit(const wordlist *args)
{
    passwdObject *m = new(passwdObject);

    m->chain.typ = FIL_AUTH;
    m->info.version = MODULE_API_VERSION;
    m->description = "password authenticator";
    m->filter = auth_passwdFilter;

    if (args) {
        m->fileName = leakAdd(xstrdup(args->key));
        args = args->next;
    } else {
        m->fileName = leakAdd(xstrdup("/etc/passwd"));
    }
    if (args) {
        m->sizefactor = atof(args->key);
        if (m->sizefactor < 0.0)
            m->sizefactor = 0.0;
    } else {
        m->sizefactor = 0.6;
    }

    moduleRegister((moduleObject *)m);
}

